// Author 		: F. de la Parra
// Date			: March, 2019
// Description  : This program executes short demos of concepts in Java 'Generics'.
//				  It uses the generic classes: ArrayList<T> and Comparison<T>.
// Version      : 1.0

import java.util.Scanner;
import java.util.ArrayList;
import java.util.InputMismatchException;


public class GenericsDemo {
	// Enumeration to label tests. 
	enum demo {NONE, ARRAYLIST, COMPARISON, DEMO_C, EXIT};
	
	public static void main(String[] args) {
		 
		// Create a 'GenericsDemo' object.
		GenericsDemo oneInstance = new GenericsDemo();
		
		// Create Scanner object 'menuSelect' for input from keyboard. 
		Scanner menuSelect = new Scanner(System.in);
		
		demo selector = demo.NONE;
		int menuItem = 0;
		
		while (!selector.equals(demo.EXIT)) {
			// Display menu
			System.out.println("\n*** Generics Demo ***");
			System.out.println("1 - ArrayList<T>");
			System.out.println("2 - Comparison<T>");
			System.out.println("9 - Exit");
			System.out.print("> ");
			
			// Input a menu selection.
			try {
				menuItem = menuSelect.nextInt();
				
				selector = oneInstance.toDemoEnum(menuItem);
				if (!selector.equals(demo.EXIT))
					System.out.println("\n*** " + selector + " test selected ***\n");
				
				switch (selector) {
						
					case ARRAYLIST:
						oneInstance.runArrayList();
						break;
						
					case COMPARISON:
						oneInstance.runComparison();
						break;
				}
			}
			catch(InputMismatchException e) {
				System.out.println("Invalid selection. Try again!");
				menuSelect.next();
				selector = demo.NONE;
			}
		}
	}
	
	// Methods for running specific tests on 'generics'.
	
	// Tests on the 'ArrayList<T>' class.
	private void runArrayList() {
		// Methods of ArrayList: get, add, remove, removeRange, clear, indexOf, isEmpty,
		// 						 ensureCapacity, size, toArray, clone, equals.
		
		ArrayList<String> names = new ArrayList<String>(); // Initial capacity of 10.
		System.out.println("Size of 'names' ArrayList = " + names.size());		
	
		// Add to the end of ArrayList<String>.
		names.add("John");
		names.add("Lisa");
		names.add("Mary");
		
		// Add at index location 1.
		names.add(1, "Andrew");
		System.out.println("Second name on the list of names: " + names.get(1));
		
		// Remove from ArrayList<String> at index location 0.
		names.remove(0);

		// Indices of objects in ArrayList<String> are automatically re-arranged.
		System.out.println("Second name on the list of names: " + names.get(1) + "\n");	
		
		// For each iteration on entire ArrayList<String>.
		System.out.println("All names:");
		for (String item : names)
			System.out.println(item);
		
		// Use "Applicant" as the base class for ArrayList<Applicant>.
		ArrayList<Applicant> admission = new ArrayList<Applicant>(100); // Initial capacity of 100.
		System.out.println("\nSize of 'admission' ArrayList = " + admission.size());
		
		// Add Applicant objects using default constructor or anonymous objects.
		admission.add(new Applicant());
		admission.add(new Applicant(123456, 85, "Patrick Doe", "Analyst"));
		admission.add(new Applicant(13579, 96, "David Hume", "Philosopher"));
		admission.add(new Applicant(124816, 99, "John Locke", "Philosopher"));
		
		// Display the toString() override version of the "Applicant" class.
		System.out.println("\n" + admission.get(0).toString());
		System.out.println("\n" + admission.get(1).toString());
		
		// Update and display and Applicant object. Use of an anonymous object
		admission.set(0, new Applicant(112233, 92, "Jordan Peterson", "Psychologist"));
		System.out.println("\n" + admission.get(0).toString());
		
	}
	
	// Tests Comparison<T> generic class.
	private void runComparison(){
		// Use of a "Comparison" custom made generic class.
		
		// Create comparison object using default constructor.
		System.out.println("*** Comparison of objects created with default constructor ***");
		Comparison<Applicant> compareOp1 = new Comparison<Applicant>();
		System.out.println("Default Comparison<Applicant> object: " + compareOp1.toString());
		
		// Create comparison object using full parameter constructor.
		System.out.println("\n*** Comparison of objects created with full parameter constructor ***");
		Comparison<Applicant> compareOp2 = new Comparison<Applicant>
											  (new Applicant(13579, 96, "David Hume", "Philosopher"),
											   new Applicant(124816, 99, "John Locke", "Philosopher"));
		
		// Display the elements of the comparison object.
		System.out.println("Attributes of Comparison<Applicant> object: \n" + compareOp2.toString());
		
		// Compare two anonymous objects with the same data.
		System.out.println("\n*** Comparing two anonymous objects with same attribute values ***");
		System.out.println("\nComparison result: " + 
						   compareOp2.compareEqual(new Applicant(13579, 96, "David Hume", "Philosopher"),
												   new Applicant(13579, 96, "David Hume", "Philosopher")));
		
		// Compare for equality using the same object.
		System.out.println("\n*** Comparing the same object ***");
		Applicant John = new Applicant(124816, 99, "John Locke", "Philosopher");
		System.out.println("\nComparison result: " + compareOp2.compareEqual(John, John));
		
		// One of the objects is 'null'. 
		System.out.println("\n*** One of the objects in the comparison is null (compareEqual) ***");
		System.out.println("\nComparison result: " + compareOp2.compareEqual(John, null));
		

		
		// Using compareTo (specified by the Comparable interface).
		System.out.println("\n*** Compare ordering using override of compareTo in Comparable interface ***");
		Applicant David = new Applicant(13579, 96, "David Hume", "Philosopher");
		System.out.println("\nComparison result: " + compareOp2.compareOrder(John, David));
		
		// Using compareTo when one of the objects is 'null'. 
		System.out.println("\n*** One of the objects in the comparison is null (compareTo) ***");
		System.out.println("\nComparison result: " + compareOp2.compareOrder(John, null));
	}
	
	// Auxiliary methods.
	
	// Convert integer input selection to enumeration objects.	
	private demo toDemoEnum(int selection) {
		demo selected = demo.NONE;
		if (selection == 1) 
			selected = demo.ARRAYLIST;
		else if (selection == 2)
			selected = demo.COMPARISON;
		else if (selection == 9)
			selected = demo.EXIT;
		return selected;
	}
} 