// Author 		: F. de la Parra
// Date			: March, 2019
// Description  : This program runs a series of demos about the features
//				  of the Java lambda expressions and their application.

import java.util.Scanner;
import java.util.InputMismatchException;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.stream.Stream;
import java.util.Collections;


public class LambdaDemo {
	// Enumeration to label the demos. 
	enum demo {NONE, SORT, SPECIAL, RUN, STREAM, D5, D6, D7, EXIT};
	
	public static void main(String[] args) {
		
		// Create a 'LambdaDemo' object.
		LambdaDemo oneInstance = new LambdaDemo();
		
		// Create Scanner object 'menuSelect' for input from keyboard. 
		Scanner menuSelect = new Scanner(System.in);
		
		// Initialize demo selector.
		demo selector = demo.NONE;
		int menuItem = 0;
		
		while (!selector.equals(demo.EXIT)) {
			// Display menu
			System.out.println("\n*** Lambda Expressions Demos ***");
			System.out.println("1 - Use in Sort");
			System.out.println("2 - Special Cases");
			System.out.println("3 - Use with Runnable Interface");
			System.out.println("4 - Use in Streams");
			System.out.println("9 - Exit");
			System.out.print("> ");
			
			// Input a menu selection.
			try {
				menuItem = menuSelect.nextInt();
				
				selector = oneInstance.toDemoEnum(menuItem);
				if (!selector.equals(demo.EXIT))
					System.out.println("\n" +selector + " demo selected\n");
				
				switch (selector) {
					case SORT:
						// Check basic methods in object
						oneInstance.sortDemo();
						break;

					case SPECIAL:
						// Check basic methods in object
						oneInstance.specialCasesDemo();
						break;
					case RUN:
						// Check basic methods in object
						oneInstance.runDemo();
						break;
					case STREAM:
						// Check basic methods in object
						oneInstance.streamDemo();
						break;								
				}
			}
			catch(InputMismatchException e) {
				System.out.println("Invalid selection. Try again!");
				menuSelect.next();
				selector = demo.NONE;
			}
			System.out.println("\nDone.");
		}
		
	}
	
	// DEMO methods
	
	// Demo use of lambda expressions in sorting a collection.
	private void sortDemo() {
		
		System.out.println("Input array: " + 
			"{'There', 'is', 'nothing', 'permanent', 'except', 'change'}\n");
		
		// VERSION-1 of lambda expression. (Example: compare objects when sorting an array).
		// Parameter types in lambda expression are explicit. It implements the 'compare'
		// abstract method of the 'Comparator' interface.
		String[] words1 = {"There", "is", "nothing", "except", "permanent", "change"};
		
		Arrays.sort(words1, (String m, String n) -> Integer.compare(m.length(), n.length()));
		
		System.out.println("Case 1: parameter types in sorting lambda expression are explicit.");
		System.out.println(Arrays.toString(words1));
		
		// VERSION-2 of lambda expression. (Example: compare objects when sorting an array).
		// Parameter types in lambda expression are inferred. It implements the 'compare'
		// abstract method of the 'Comparator' functional interface.
		String[] words2 = {"There", "is", "nothing", "permanent", "except", "change"};
		
		Arrays.sort(words2, (m, n) -> Integer.compare(m.length(), n.length()));
		
		System.out.println("\nCase 2: inferred parameter types in sorting lambda expression.");
		System.out.println(Arrays.toString(words2));

		// VERSION-3 of lambda expression. (Example: compare objects when sorting an array).
		// Multiple lines in lambda expression and redefinition of comparison. It implements
		// the 'compare' abstract method of the 'Comparator' functional interface.	
		String[] words3 = {"There", "is", "nothing", "permanent", "except", "change"};
		
		Arrays.sort(words3, (String m, String n) -> {
					if (m.length() > n.length())
						return -1;
					else
						return 0;
					});
		System.out.println("\nCase 3: multiple-line lambda expression redefines sort order.");
		System.out.println(Arrays.toString(words3));
		
	}
	
	// Demo use of minimal and composite lambda expressions.
	private void specialCasesDemo() {
		
		System.out.println("Input array: " + 
			"{'There', 'is', 'nothing', 'permanent', 'except', 'change'}\n");
		
		// CASE 1:
		// Treat the elements of an array as a stream of tokens and operate on them.
		// Lambda expression type: single parameter, no round brackets, inferred type.
		// It implements the 'accept' abstract method of the 'Consumer' functional
		// interface.
		String[] words = {"There", "is", "nothing", "permanent", "except", "change"};
		Stream<String> myTokenizer1 = Stream.of(words);
		System.out.println("Case 1: Single inferred parameter in lambda expression.");
		
		myTokenizer1.forEach(x -> System.out.print(x.toUpperCase() + " "));	
		
		
		// CASE 2:
		// Lambda expression type: No parameters. It implements the 'doSomething'
		// abstract method of the 'BasicInterface' functional interface (user-defined).
		
		BasicInterface printObj = () -> {System.out.println(
								 "<<< Word 'change' found in parsed sentence >>>");};
		
		// Reuse 'printObj' object generated with the previous lambda expression in
		// a new one that implements the 'accept' abstract method of the 'Consumer'
		// functional interface.
		Stream<String> myTokenizer2 = Stream.of(words);
		System.out.println("\n\nCase 2: No parameters in lambda expression.");
		
		myTokenizer2.forEach(x -> {if (x.equals("change")) printObj.doSomething();});
	}
	
	// Demo use of runnable interface.
	private void runDemo() {
		Runnable r = () -> System.out.println("Printed from a lambda expression!");
		r.run();
	}

	// Demo use of runnable interface.
	private void streamDemo() {
		// Create an array list of applcants.
		ArrayList<Applicant> applicants = new ArrayList<Applicant>();
			applicants.add(new Applicant(123456, 85, "Patrick Doe", "Analyst"));
			applicants.add(new Applicant(13579, 96, "David Hume", "Philosopher"));
			applicants.add(new Applicant(124816, 99, "John Locke", "Philosopher"));
			applicants.add(new Applicant(112233, 92, "Jordan Peterson", "Psychologist"));
			applicants.add(new Applicant(112233, 97, "Alan Turing", "Mathematician"));
		
		// Select and display the applicants with a score > 90.
		System.out.println("*** Print stream of selected applicants ***");
		applicants.stream()
				.filter(a -> a.getScore() > 90)
				.map(s -> s.getFullName() + "\t\t" + s.getScore())
				.forEach(s -> System.out.println(s));
		
		// Display the list of selected applicants in descending score order.
		System.out.println("\n*** Print stream of selected and sorted applicants ***");
		Collections.sort(applicants, (Applicant a, Applicant b) -> {
					if (a.getScore() > b.getScore())
						return -1;
					else
						return 0;
					});
					
		applicants.stream()
				.filter(a -> a.getScore() > 90)
				.map(s -> s.getFullName() + "\t\t" + s.getScore())
				.forEach(s -> System.out.println(s));
	}	

	// Auxiliary methods.
	
	private demo toDemoEnum(int selection) {
		demo selected = demo.NONE;
		if (selection == 1) 
			selected = demo.SORT;
		else if (selection == 2)
			selected = demo.SPECIAL;
		else if (selection == 3)
			selected = demo.RUN;
		else if (selection == 4)
			selected = demo.STREAM;
		else if (selection == 9)
			selected = demo.EXIT;
		return selected;
	}
	
}